@@ -1,7 +1,7 @@
 /*
  /  Author: Sam Rushing <rushing@nightmare.com>
  /  Hacked for Unix by AMK
- /  $Id: mmapmodule.c,v 2.35.6.5 2003/02/07 19:46:44 nnorwitz Exp $
+ /  $Id: mmapmodule.c,v 2.46.6.1 2004/06/01 13:00:16 akuchling Exp $
 
  / mmapmodule.cpp -- map a view of a file into memory
  /
@@ -20,11 +20,11 @@
 
 #include <Python.h>
 
-#ifndef MS_WIN32
+#ifndef MS_WINDOWS
 #define UNIX
 #endif
 
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 #include <windows.h>
 static int
 my_getpagesize(void)
@@ -75,7 +75,7 @@ typedef struct {
 	size_t	size;
 	size_t	pos;
 
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 	HANDLE	map_handle;
 	HANDLE	file_handle;
 	char *	tagname;
@@ -92,7 +92,7 @@ typedef struct {
 static void
 mmap_object_dealloc(mmap_object *m_obj)
 {
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 	if (m_obj->data != NULL)
 		UnmapViewOfFile (m_obj->data);
 	if (m_obj->map_handle != INVALID_HANDLE_VALUE)
@@ -101,7 +101,7 @@ mmap_object_dealloc(mmap_object *m_obj)
 		CloseHandle (m_obj->file_handle);
 	if (m_obj->tagname)
 		PyMem_Free(m_obj->tagname);
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 #ifdef UNIX
 	if (m_obj->data!=NULL) {
@@ -118,7 +118,7 @@ mmap_close_method(mmap_object *self, PyObject *args)
 {
         if (!PyArg_ParseTuple(args, ":close"))
 		return NULL;
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 	/* For each resource we maintain, we need to check
 	   the value is valid, and if so, free the resource 
 	   and set the member value to an invalid value so
@@ -138,7 +138,7 @@ mmap_close_method(mmap_object *self, PyObject *args)
 		CloseHandle (self->file_handle);
 		self->file_handle = INVALID_HANDLE_VALUE;
 	}
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 #ifdef UNIX
 	if (self->data != NULL) {
@@ -151,15 +151,15 @@ mmap_close_method(mmap_object *self, PyObject *args)
 	return (Py_None);
 }
 
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 #define CHECK_VALID(err)						\
 do {									\
-    if (!self->map_handle) {						\
+    if (self->map_handle == INVALID_HANDLE_VALUE) {						\
 	PyErr_SetString (PyExc_ValueError, "mmap closed or invalid");	\
 	return err;							\
     }									\
 } while (0)
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 #ifdef UNIX
 #define CHECK_VALID(err)						\
@@ -338,7 +338,7 @@ mmap_size_method(mmap_object *self,
         if (!PyArg_ParseTuple(args, ":size"))
 		return NULL;
 
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 	if (self->file_handle != INVALID_HANDLE_VALUE) {
 		return (Py_BuildValue (
 			"l", (long)
@@ -346,7 +346,7 @@ mmap_size_method(mmap_object *self,
 	} else {
 		return (Py_BuildValue ("l", (long) self->size) );
 	}
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 #ifdef UNIX
 	{
@@ -378,7 +378,7 @@ mmap_resize_method(mmap_object *self,
 	if (!PyArg_ParseTuple (args, "l:resize", &new_size) || 
 	    !is_resizeable(self)) {
 		return NULL;
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 	} else { 
 		DWORD dwErrCode = 0;
 		/* First, unmap the file view */
@@ -416,7 +416,7 @@ mmap_resize_method(mmap_object *self,
 		}
 		PyErr_SetFromWindowsErr(dwErrCode);
 		return (NULL);
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 #ifdef UNIX
 #ifndef HAVE_MREMAP 
@@ -469,10 +469,10 @@ mmap_flush_method(mmap_object *self, PyObject *args)
 				 "flush values out of range");
 		return NULL;
 	} else {
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 		return (Py_BuildValue("l", (long)
                                       FlushViewOfFile(self->data+offset, size)));
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 #ifdef UNIX
 		/* XXX semantics of return value? */
 		/* XXX flags for msync? */
@@ -556,19 +556,19 @@ mmap_move_method(mmap_object *self, PyObject *args)
 }
 
 static struct PyMethodDef mmap_object_methods[] = {
-	{"close",	(PyCFunction) mmap_close_method,	1},
-	{"find",	(PyCFunction) mmap_find_method,		1},
-	{"flush",	(PyCFunction) mmap_flush_method,	1},
-	{"move",	(PyCFunction) mmap_move_method,		1},
-	{"read",	(PyCFunction) mmap_read_method,		1},
-	{"read_byte",	(PyCFunction) mmap_read_byte_method,  	1},
-	{"readline",	(PyCFunction) mmap_read_line_method,	1},
-	{"resize",	(PyCFunction) mmap_resize_method,	1},
-	{"seek",	(PyCFunction) mmap_seek_method,		1},
-	{"size",	(PyCFunction) mmap_size_method,		1},
-	{"tell",	(PyCFunction) mmap_tell_method,		1},
-	{"write",	(PyCFunction) mmap_write_method,	1},
-	{"write_byte",	(PyCFunction) mmap_write_byte_method,	1},
+	{"close",	(PyCFunction) mmap_close_method,	METH_VARARGS},
+	{"find",	(PyCFunction) mmap_find_method,		METH_VARARGS},
+	{"flush",	(PyCFunction) mmap_flush_method,	METH_VARARGS},
+	{"move",	(PyCFunction) mmap_move_method,		METH_VARARGS},
+	{"read",	(PyCFunction) mmap_read_method,		METH_VARARGS},
+	{"read_byte",	(PyCFunction) mmap_read_byte_method,  	METH_VARARGS},
+	{"readline",	(PyCFunction) mmap_read_line_method,	METH_VARARGS},
+	{"resize",	(PyCFunction) mmap_resize_method,	METH_VARARGS},
+	{"seek",	(PyCFunction) mmap_seek_method,		METH_VARARGS},
+	{"size",	(PyCFunction) mmap_size_method,		METH_VARARGS},
+	{"tell",	(PyCFunction) mmap_tell_method,		METH_VARARGS},
+	{"write",	(PyCFunction) mmap_write_method,	METH_VARARGS},
+	{"write_byte",	(PyCFunction) mmap_write_byte_method,	METH_VARARGS},
 	{NULL,	   NULL}       /* sentinel */
 };
 
@@ -897,7 +897,12 @@ new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 	}
 
 #ifdef HAVE_FSTAT
-	if (fstat(fd, &st) == 0 && (size_t)map_size > st.st_size) {
+#  ifdef __VMS
+	/* on OpenVMS we must ensure that all bytes are written to the file */
+	fsync(fd);
+#  endif
+	if (fstat(fd, &st) == 0 && S_ISREG(st.st_mode) &&
+	    (size_t)map_size > st.st_size) {
 		PyErr_SetString(PyExc_ValueError, 
 				"mmap length is greater than file size");
 		return NULL;
@@ -912,6 +917,7 @@ new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 			   prot, flags,
 			   fd, 0);
 	if (m_obj->data == (char *)-1) {
+	        m_obj->data = NULL;
 		Py_DECREF(m_obj);
 		PyErr_SetFromErrno(mmap_module_error);
 		return NULL;
@@ -921,7 +927,7 @@ new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 }
 #endif /* UNIX */
 
-#ifdef MS_WIN32
+#ifdef MS_WINDOWS
 static PyObject *
 new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 {
@@ -974,7 +980,7 @@ new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 			return NULL;
 		}
 		/* Win9x appears to need us seeked to zero */
-		fseek(&_iob[fileno], 0, SEEK_SET);
+		lseek(fileno, 0, SEEK_SET);
 	}
 
 	m_obj = PyObject_New (mmap_object, &mmap_object_type);
@@ -1054,7 +1060,7 @@ new_mmap_object(PyObject *self, PyObject *args, PyObject *kwdict)
 	PyErr_SetFromWindowsErr(dwErr);
 	return (NULL);
 }
-#endif /* MS_WIN32 */
+#endif /* MS_WINDOWS */
 
 /* List of functions exported by this module */
 static struct PyMethodDef mmap_functions[] = {
@@ -1063,7 +1069,7 @@ static struct PyMethodDef mmap_functions[] = {
 	{NULL,		NULL}	     /* Sentinel */
 };
 
-DL_EXPORT(void)
+PyMODINIT_FUNC
 	initmmap(void)
 {
 	PyObject *dict, *module;
